#include <iostream>

// LibXML2 Includes
extern "C" {
#include <libxml/parser.h>
#include <libxml/tree.h>
}

using namespace std;

// Class to visit a method body text
class BodyText {
private:
    // XML Element to visit
    xmlNodePtr m_node;
    // Indentation level
    int m_tabLevel;
public:
    // Constructor for 'bodytext' object
    BodyText(xmlNodePtr node, int tabLevel = 0):
        m_node(node),
        m_tabLevel(tabLevel){
        // Check if we are dealing with a 'text' element
        if(xmlStrcmp(m_node->name,
                     (const xmlChar*) "text")){
            // Bad instantiation, throw exception string
            throw string("XML-Element/Visitor-Class mismatch");
        }
    }
    // Method to get 'bodytext'-element text
    string getContent() const{
        // Return content as string
        return string((const char*) m_node->content);
    }
    // Method to visit a 'bodytext' element
    void visit() const{
        // Output indentation
        for(int i = 0; i < m_tabLevel; i++) { cout << "   "; }
        // Output body text
        cout << getContent()
             << endl;
    }
};

// Class to visit a method body
class Body {
private:
    // XML Element to visit
    xmlNodePtr m_node;
    int m_tabLevel;
public:
    // Constructor for 'body' object
    Body(xmlNodePtr node, int tabLevel = 0):
        m_node(node),
        m_tabLevel(tabLevel){
        // Check if we are dealing with a 'return' element
        if(xmlStrcmp(m_node->name,
                     (const xmlChar*) "body")){
            // Bad instantiation, throw exception string
            throw string("XML-Element/Visitor-Class mismatch");
        }
    }
    // Method to visit a 'body' element
    void visit() const{
        // Visit child 'text' elements
        {
            // Loop over children
            for(xmlNodePtr child = m_node->children;
                child;
                child = child->next){
                // Check if current child is a 'text' element
                if(not xmlStrcmp(child->name,
                                 (const xmlChar*) "text")){
                    // Instantiate 'text' object
                    BodyText bodytext(child, m_tabLevel);
                    // Visit 'text' object
                    bodytext.visit();
                }
            }
        }

    }
};

// Class to visit a method return type
class Return {
private:
    // XML Element to visit
    xmlNodePtr m_node;
    // Indentation level
    int m_tabLevel;
public:
    // Constructor for 'return' object
    Return(xmlNodePtr node, int tabLevel = 0):
        m_node(node),
        m_tabLevel(tabLevel){
        // Check if we are dealing with a 'return' element
        if(xmlStrcmp(m_node->name,
                     (const xmlChar*) "return")){
            // Bad instantiation, throw exception string
            throw string("XML-Element/Visitor-Class mismatch");
        }
    }
    // Method to get 'return'-element attribute 'type'
    string getType() const{
        // Read 'type' attribute
        xmlChar *type = xmlGetProp(m_node, (const xmlChar*) "type");
        // Check if attribute exists
        if(not type){ throw string("Return element without type attribute"); }
        // Construct return string
        string ret((const char*) type);
        // Free attribute buffer
        xmlFree(type);
        // return attribute value
        return ret;
    }
    // Method to visit a 'return' element
    void visit() const{
        // Output return type
        cout << getType();
    }
};

// Class to visit a method formal parameter
class Parameter {
private:
    // XML Element to visit
    xmlNodePtr m_node;
    // Indentation level
    int m_tabLevel;
public:
    // Constructor for 'parameter' object
    Parameter(xmlNodePtr node, int tabLevel = 0):
        m_node(node),
        m_tabLevel(tabLevel){
        // Check if we are dealing with a 'method' element
        if(xmlStrcmp(m_node->name,
                     (const xmlChar*) "parameter")){
            // Bad instantiation, throw exception string
            throw string("XML-Element/Visitor-Class mismatch");
        }
    }
    // Method to get 'parameter'-element attribute 'name'
    string getName() const{
        // Read 'name' attribute
        xmlChar *name = xmlGetProp(m_node, (const xmlChar*) "name");
        // Check if attribute exists
        if(not name){ throw string("Parameter element without name attribute"); }
        // Construct return string
        string ret((const char*) name);
        // Free attribute buffer
        xmlFree(name);
        // return attribute value
        return ret;
    }
    // Method to get 'parameter'-element attribute 'type'
    string getType() const{
        // Read 'type' attribute
        xmlChar *type = xmlGetProp(m_node, (const xmlChar*) "type");
        // Check if attribute exists
        if(not type){ throw string("Parameter element without type attribute"); }
        // Construct return string
        string ret((const char*) type);
        // Free attribute buffer
        xmlFree(type);
        // return attribute value
        return ret;
    }
    // Method to visit a 'parameter' element
    void visit() const{
        // Output parameter declaration
        cout << getType()
             << " "
             << getName();
    }
};

// Class to visit a method
class Method {
private:
    // XML Element to visit
    xmlNodePtr m_node;
    // Indentation level
    int m_tabLevel;
public:
    // Constructor for 'method' object
    Method(xmlNodePtr node, int tabLevel = 0):
        m_node(node),
        m_tabLevel(tabLevel){
        // Check if we are dealing with a 'method' element
        if(xmlStrcmp(m_node->name,
                     (const xmlChar*) "method")){
            // Bad instantiation, throw exception string
            throw string("XML-Element/Visitor-Class mismatch");
        }
    }
    // Method to get 'method'-element attribute 'name'
    string getName() const{
        // Read 'name' attribute
        xmlChar *name = xmlGetProp(m_node, (const xmlChar*) "name");
        // Check if attribute exists
        if(not name){ throw string("Method element without name attribute"); }
        // Construct return string
        string ret((const char*) name);
        // Free attribute buffer
        xmlFree(name);
        // return attribute value
        return ret;
    }
    // Method to visit a 'method' element
    void visit() const{
        // Output indentation
        for(int i = 0; i < m_tabLevel; i++) { cout << "   "; }
        // Visit child 'return' element
        {
            // Loop over children
            for(xmlNodePtr child = m_node->children;
                child;
                child = child->next){
                // Check if current child is a 'return' element
                if(not xmlStrcmp(child->name,
                                 (const xmlChar*) "return")){
                    // Instantiate 'return' object
                    Return ret(child, m_tabLevel + 1);
                    // Visit 'return' object
                    ret.visit();
                    // Break after first
                    break;
                }
            }
        }
        // Output method name
        cout << " " << getName();
        // Output formal parameter list opener
        cout << "(";
        // Visit child 'parameter' elements
        {
            // Loop over children
            for(xmlNodePtr child = m_node->children;
                child;
                child = child->next){
                // Check if current child is a 'parameter' element
                if(not xmlStrcmp(child->name,
                                 (const xmlChar*) "parameter")){
                    // Instantiate 'parameter' object
                    Parameter param(child, m_tabLevel + 1);
                    // Visit 'parameter' object
                    param.visit();
                    // Output ", " if another parameter is expected
                    if((child->next) and
                       (not xmlStrcmp(child->next->name,
                                      (const xmlChar*) "parameter"))) { cout << ", "; }
                }
            }
        }
        // Output formal parameter list closer
        cout << ")";
        // Output body opener
        cout << "{" << endl;
        // Visit child 'body' element
        {
            // Loop over children
            for(xmlNodePtr child = m_node->children;
                child;
                child = child->next){
                // Check if current child is a 'body' element
                if(not xmlStrcmp(child->name,
                                 (const xmlChar*) "body")){
                    // Instantiate 'body' object
                    Body body(child, m_tabLevel + 1);
                    // Visit 'body' object
                    body.visit();
                    // Break upon body
                    break;
                }
            }
        }
        // Output indentation
        for(int i = 0; i < m_tabLevel; i++) { cout << "   "; }
        // Output body closer
        cout << "}" << endl;
    }
};

// Class to visit a class attribute
class Attribute {
private:
    // XML Element to visit
    xmlNodePtr m_node;
    // Indentation level
    int m_tabLevel;
public:
    // Constructor for 'attribute' object
    Attribute(xmlNodePtr node, int tabLevel = 0):
        m_node(node),
        m_tabLevel(tabLevel){
        // Check if we are dealing with an 'attribute' element
        if(xmlStrcmp(m_node->name,
                     (const xmlChar*) "attribute")){
            // Bad instantiation, throw exception string
            throw string("XML-Element/Visitor-Class mismatch");
        }
    }
    // Method to get 'attribute'-element attribute 'name'
    string getName() const{
        // Read 'name' attribute
        xmlChar *name = xmlGetProp(m_node, (const xmlChar*) "name");
        // Check if attribute exists
        if(not name){ throw string("Attribute element without name attribute"); }
        // Construct return string
        string ret((const char*) name);
        // Free attribute buffer
        xmlFree(name);
        // return attribute value
        return ret;
    }
    // Method to get 'attribute'-element attribute 'type'
    string getType() const{
        // Read 'type' attribute
        xmlChar *type = xmlGetProp(m_node, (const xmlChar*) "type");
        // Check if attribute exists
        if(not type){ throw string("Attribute element without type attribute"); }
        // Construct return string
        string ret((const char*) type);
        // Free attribute buffer
        xmlFree(type);
        // return attribute value
        return ret;
    }
    // Method to visit an 'attribute' element
    void visit() const{
        // Output indentation
        for(int i = 0; i < m_tabLevel; i++) { cout << "   "; }
        // Output attribute declaration
        cout << getType()
             << " "
             << getName()
             << ";"
             << endl;
    }
};

// Class to visit a class element
class Class {
private:
    // XML Element to visit
    xmlNodePtr m_node;
    // Indentation level
    int m_tabLevel;
public:
    // Constructor for 'class' object
    Class(xmlNodePtr node, int tabLevel = 0):
        m_node(node),
        m_tabLevel(tabLevel){
        // Check if we are dealing with a 'class' element
        if(xmlStrcmp(m_node->name,
                     (const xmlChar*) "class")){
            // Bad instantiation, throw exception string
            throw string("XML-Element/Visitor-Class mismatch");
        }
    }
    // Method to get 'class'-element attribute 'name'
    string getName() const{
        // Read 'name' attribute
        xmlChar *name = xmlGetProp(m_node, (const xmlChar*) "name");
        // Check if attribute exists
        if(not name){ throw string("Class element without name attribute"); }
        // Construct return string
        string ret((const char*) name);
        // Free attribute buffer
        xmlFree(name);
        // return attribute value
        return ret;
    }
    // Method to visit a 'class' element
    void visit() const{
        // Output indentation
        for(int i = 0; i < m_tabLevel; i++) { cout << "   "; }
        // Output class opener
        cout << "class "
             << getName() 
             << " {"
             << endl;
        // Visit child 'attribute' elements
        {
            // Output indentation
            for(int i = 0; i < m_tabLevel; i++) { cout << "   "; }
            // Output visibility
            cout << " private:" << endl;
            // Loop over children
            for(xmlNodePtr child = m_node->children;
                child;
                child = child->next){
                // Check if current child is a 'attribute' element
                if(not xmlStrcmp(child->name,
                                 (const xmlChar*) "attribute")){
                    // Instantiate 'attribute' object
                    Attribute attr(child, m_tabLevel + 1);
                    // Visit 'attribute' object
                    attr.visit();
                }
            }
        }
        // Visit child 'method' elements
        {
            // Output indentation
            for(int i = 0; i < m_tabLevel; i++) { cout << "   "; }
            // Output visibility
            cout << " public:" << endl;
            // Loop over children
            for(xmlNodePtr child = m_node->children;
                child;
                child = child->next){
                // Check if current child is a 'method' element
                if(not xmlStrcmp(child->name,
                                 (const xmlChar*) "method")){
                    // Instantiate 'method' object
                    Method method(child, m_tabLevel + 1);
                    // Visit 'method' object
                    method.visit();
                }
            }
        }
        // Output indentation
        for(int i = 0; i < m_tabLevel; i++) { cout << "   "; }
        // Output class closer
        cout << "};" << endl;
    }
};

// Class to visit a model
class Model {
private:
    // XML Element to visit
    xmlNodePtr m_node;
    // Indentation level
    int m_tabLevel;
public:
    // Constructor for 'model' object
    Model(xmlNodePtr node, int tabLevel = 0):
        m_node(node),
        m_tabLevel(tabLevel){
        // Check if we are dealing with a 'model' element
        if(xmlStrcmp(m_node->name,
                     (const xmlChar*) "model")){
            // Bad instantiation, throw exception string
            throw string("XML-Element/Visitor-Class mismatch");
        }
    }
    // Method to visit a 'model' element
    void visit() const{
        // Visit child 'class' elements
        {
            // Loop over child elements
            for(xmlNodePtr child = m_node->children;
                child;
                child = child->next){
                // Check if current child is a 'class' element
                if(not xmlStrcmp(child->name,
                                 (const xmlChar*) "class")){
                    // Instantiate 'child' object
                    Class cls(child, 0);
                    // Visit 'child' object
                    cls.visit();
                }
            }
        }
    }
};


// Main routine
int main(int argc, char* argv[]){

    // Runtime check for correct LibXML2 version
    LIBXML_TEST_VERSION

    if(argc != 2){
        cerr << "Usage: example <FILEPATH>" << endl;
        return 1;
    }

    // Parse input file
    xmlDocPtr domRoot = xmlReadFile(argv[1], NULL, 0);

    // Check if parsing was successful
    if(not domRoot){
        cerr << "FAILED: Bad input file" << endl;
        return 1;
    }

    // Recover root node
    xmlNodePtr rootNode = xmlDocGetRootElement(domRoot);

    // Attempt to perform passes on DOM
    try{
        
        // Instantiate 'model' object
        Model model(rootNode);
        // Visit 'model' object
        model.visit();

    }
    // Catch exception string
    catch(const string &exception){
        cerr << "FAILED: " << exception << endl;
        return 1;
    }

    // Free document
    xmlFreeDoc(domRoot);

    return 0;
}
